PROGRAM example2
USE tensolve
IMPLICIT NONE

!   TENSOLVE finds roots of systems of n nonlinear equations in
!   n unknowns, or minimizers of the sum of squares of m > n
!   nonlinear functions in n unknowns, using tensor methods.

!   This example illustrates the use of TENSOLVE to solve a nonlinear
!   least-squares problem defined by subroutines fwood and jwood
!   (included below).

! N.B. The original, Fortran 77 version, used method = 1, global = 1

INTEGER            :: m, n, itnlim, jacflg, method, global, ipr, msg, termcd
INTEGER            :: meth, glob
REAL (dp)          :: gradtl, steptl, ftol, stepmx, dlt
INTEGER, PARAMETER :: maxm = 100, maxn = 30, maxp = 6
REAL (dp)          :: x0(maxn), xp(maxn), fp(maxm), gp(maxn), typx(maxn), &
                      typf(maxm)

INTERFACE
  SUBROUTINE fwood ( x, f, m, n )
    USE tensolve, ONLY: dp
    IMPLICIT NONE
    REAL (dp), INTENT(IN)   :: x(:)
    REAL (dp), INTENT(OUT)  :: f(:)
    INTEGER, INTENT(IN)     :: m
    INTEGER, INTENT(IN)     :: n
  END SUBROUTINE fwood

  SUBROUTINE jwood ( x, jac, m, n )
    USE tensolve, ONLY: dp
    IMPLICIT NONE
    REAL (dp), INTENT(IN)   :: x(:)
    REAL (dp), INTENT(OUT)  :: jac(:,:)
    INTEGER, INTENT(IN)     :: m
    INTEGER, INTENT(IN)     :: n
  END SUBROUTINE jwood
END INTERFACE

!     Set dimensions of the problem.

m      = 6
n      = 4

DO meth = 0, 1
  DO glob = 0, 1

!     Set values for the initial point.

    x0(1)  = -300.0_dp
    x0(2)  = -100.0_dp
    x0(3)  = -300.0_dp
    x0(4)  = -100.0_dp

!     Set default values for the TENSOLVE parameters.

    CALL tsdflt( m, n, itnlim, jacflg, gradtl, steptl,  &
                 ftol, method, global, stepmx, dlt, typx, typf, ipr)

!     Alter some of the parameters.

    gradtl = 1.0e-5_dp
    ftol   = 1.0e-9_dp
    steptl = 1.0e-9_dp
    global = glob
    method = meth

    WRITE(*, '(a, i3, a, i3)') ' ** Method =', method, '   ** Global =', global

!     Call TENSOLVE.

    CALL tsneci(maxm, maxn, maxp, x0, m, n, typx, typf, itnlim, jacflg,  &
                gradtl, steptl, ftol, method, global, stepmx, dlt, ipr,  &
                fwood, jwood, msg, xp, fp, gp, termcd )

    WRITE(*, '(a, i5)')     ' Termination code = ', termcd
    WRITE(*, '(a, 4g13.5)') ' Solution: ', xp(1:4)
    WRITE(*, '(a, 6g13.5)') ' Function values at solution = ', fp(1:6)
    WRITE(*, '(a, 4g13.5)') ' Gradients at solution = ', gp(1:4)
    WRITE(*, *)
    WRITE(*, *)
  END DO
END DO

!     end of example2 main program
STOP
END PROGRAM example2



SUBROUTINE fwood ( x, f, m, n )
USE tensolve, ONLY: dp
IMPLICIT NONE

REAL (dp), INTENT(IN)       :: x(:)
REAL (dp), INTENT(OUT)      :: f(:)
INTEGER, INTENT(IN)         :: m
INTEGER, INTENT(IN)         :: n

!     fwood defines function values for the Wood function.

f(1) = 10.0_dp * (x(2) - x(1)**2)
f(2) = 1.0_dp - x(1)
f(3) = SQRT(90.0_dp) * (x(4) - x(3)**2)
f(4) = 1.0_dp - x(3)
f(5) = (x(2) + x(4) - 2.0_dp) * SQRT(10.0_dp)
f(6) = (x(2) - x(4)) / SQRT(10.0_dp)

!     end of fwood.
RETURN
END SUBROUTINE fwood



SUBROUTINE jwood ( x, jac, m, n )
USE tensolve, ONLY: dp
IMPLICIT NONE

REAL (dp), INTENT(IN)        :: x(:)
REAL (dp), INTENT(OUT)       :: jac(:,:)
INTEGER, INTENT(IN)          :: m
INTEGER, INTENT(IN)          :: n

!     jwood defines Jacobian values for the Wood function.

REAL (dp) :: tval

jac(1:m,1:n) = 0.0_dp

jac(1,1) = -20.0_dp * x(1)
jac(1,2) =  10.0_dp

jac(2,1) = -1.0_dp

tval     =  SQRT(90.0_dp)
jac(3,3) = -2.0_dp * tval * x(3)
jac(3,4) =  tval

jac(4,3) = -1.0_dp

tval     =  SQRT(10.0_dp)
jac(5,2) =  tval
jac(5,4) =  tval

tval     =  1.0_dp / tval
jac(6,2) =  tval
jac(6,4) = -tval

!     end of jwood.
RETURN
END SUBROUTINE jwood

